; TERRY COTTON ZS1AYJ
; PIC CONTROLLER FOR AmQRP DDS60 KIT
; REVISION 1.0 FIRST PUBLISHED SEPTEMBER 2006
; THIS IS REVISION 2 MAY 2013 WHICH AT SWITCH-ON
; INITIALISES TO THE FREQUENCY WHICH WAS BEING
; GENERATED AT SWITCH-OFF RATHER THAN DFAULTING TO 1 MHZ 
;**************************************************************
; Much of the code in this module is original. However, other
; parts such as the LCD & Math routines are based on code
; published in various magazine projects, particularly EPE magazine
; over the last few years. All credit to the original authors.
;**************************************************************
; 
; PIC PORT USAGE
;
; THIS PROJECT WAS GRADUALLY DEVELOPED RATHER THAN DESIGNED.
; A PIC16F876 (4 MHZ CLOCK) WAS CHOSEN AT THE OUTSET AS IT HAD
; ENOUGH PORT PINS TO ACCOMODATE WHATEVER THE PROJECT FINALLY
; TURNED OUT TO BE ! THIS IS WHAT TRANSPIRED:
;
; PORTA		RA0	UP/DOWN BUTTON
;		RA1	BEEP SOUND ON BUTTON PRESS
; 		RA2	DDS60 PIN 1 (LOAD)
; 		RA3	DDS60 PIN 2 (CLOCK)
; 		RA4	DDS60 PIN 3 (DATA)
;		RA5	UNUSED
; PORTB		RB0	LCD DATA D4
;		RB1	LCD DATA D5
;		RB2	LCD DATA D6
;		RB3	LCD DATA D7
; 		RB4	LCD RS LINE
;		RB5	LCD E LINE
;		RB6-7	FREE FOR IN-CIRCUIT PROGRAMMING
; PORTC		RC0	1 HZ BUTTON
;		RC1	10 HZ BUTTON
;		RC2	100 HZ BUTTON
;		RC3	1 KHz HZ BUTTON
;		RC4	10 KHz HZ BUTTON
;		RC5	100 KHz BUTTON
;		RC6	1 MHz BUTTON
;		RC7	10 MHz BUTTON
;**************************************************************

		include "P16F876A.inc"
		list p=16F876, r=dec, st=on ;radix 10, produce symbol table 
		;code protection off
		;power-on timer on
		;watchdog timer off
		;HS Oscillator
		;Low voltage programming off
		__config _CP_OFF & _PWRTE_ON & _WDT_OFF & _HS_OSC & _LVP_OFF

; BANK SWITCHING MACROS
BANK0		macro
		bcf STATUS,RP1  	;BANK0
        	bcf STATUS,RP0
		endm
BANK1		macro
		bcf STATUS,RP1  	;BANK1
        	bsf STATUS,RP0
		endm
BANK2		macro
		bsf STATUS,RP1  	;BANK2
        	bcf STATUS,RP0
		endm
BANK3		macro
		bsf STATUS,RP1  	;BANK3
        	bsf STATUS,RP0
		endm
;**************************************************************
; General Purpose register definitions
;**************************************************************
        	CBLOCK 020h
;**************************************************************
; REGISTERS A to C. THESE ARE USED BY 
; PETER HEMSLEY'S 32-BIT MATHS ROUTINES. FOR THIS
; APPLICATION THEY HAVE BEEN EXTENDED TO 64 BITS
; BY TERRY COTTON - ZS1AYJ
;**************************************************************

REGA0				;lsb
REGA1
REGA2
REGA3				
REGA4
REGA5
REGA6
REGA7				;msb

REGB0				;lsb
REGB1
REGB2
REGB3				
REGB4
REGB5
REGB6
REGB7				;msb

REGC0				;lsb
REGC1
REGC2
REGC3				
REGC4
REGC5
REGC6
REGC7				;msb

DSIGN				;Digit Sign. 0=positive,FF(or non zero)=negative
DIGIT1				;MSD
DIGIT2
DIGIT3
DIGIT4
DIGIT5				;Decimal digits
DIGIT6
DIGIT7
DIGIT8
DIGIT9
DIGIT10				
DIGIT11				
DIGIT12				
DIGIT13				
DIGIT14				
DIGIT15				;LSD				

MTEMP
MCOUNT
DCOUNT
OVERFLOW
POINT
;**************************************************************
; Frequency register. This maintains the frequency (in Hz)
; required by the operator. It is incremented/decremented by
; the push buttons and the result is displayed as the frequency
; on the LCD. The phase increment required for the DDS unit to 
; generate this frequency is then calculated and sent by the serial
; interface to the DDS60.
;**************************************************************
Freq0			;Freq lsd
Freq1
Freq2
Freq3			;Freq msd

LOOP                    ; loop counter 1 - general
LOOPA                   ; loop counter 2 - LCD use only
MSLOOP			; 1 mS delay routine loop
VARLOOP			; Vardelay loop
STORE                   ; store used by lcdout
RSLINE                  ; bit 4 RS line flag for LCD
DELTAHZ0		; lsb of DELTAHZ
DELTAHZ1
DELTAHZ2
DELTAHZ3		; msb of DELTAHZ

XTAL0			; The actual frequency of the 30 MHz crystal
XTAL1			; clock on the DDS60 board.This can be adjusted
XTAL2			; by the operator and is stored in EEPROM
XTAL3			; It is used to calculate the DELTAHZ value
 
TEMP			; general temp store

BEEPOUTER		; Beep routine loop values
BEEPINNER
EEPROMDATA		; Data byte for writes to EEPROM
	        ENDC
;**************************************************************

                ORG 0                   ; Reset Vector address
                GOTO 5                  ; go to PIC address location 5
                ORG 4                   ; Interrupt Vector address
                GOTO 5                  ; go to PIC address location 5
                ORG 5                   ; Start of Program Memory at location 5

 		clrf intcon		; disable all interrupts
                clrf PORTA
                clrf PORTB
                clrf PORTC
		
		BANK1
                movlw 6			; configure all porta pins
		movwf adcon1		; as digital inputs
		movlw 1			; Port A0 as input
		movwf TRISA             ; Port A1-A5 as outputs
                clrf TRISB		; PORTB as output
                movlw 0ffh
		movwf TRISC		; Port C as inputs

		movlw B'10000000'       ; move ratio value into W
                movwf OPTION_REG        ; light pull-ups off - bit 7 high
                BANK0                   

                call LCDSETUP           ; setup LCD

		b    main		; bypass tables
;**************************************************************
; LCD initialisation table
;**************************************************************
TABLCD          addwf PCL,F             ; LCD initialisation table
		retlw B'00110011'       ; initialise lcd - first byte
                retlw B'00110011'       ; 2nd byte (repeat of first)
                retlw B'00110010'       ; set for 4-bit operation
                retlw B'00101100'       ; set for 2 lines
                retlw B'00000110'       ; set entry mode to increment each address
                retlw B'00001100'       ; set display on, cursor off, blink off
                retlw B'00000001'       ; clear display
                retlw B'00000010'       ; return home, cursor & RAM to zero
		retlw 0                 ; end initialisation table
                 
;**************************************************************
; LCD SETUP ROUTINE
;**************************************************************
LCDSETUP        movlw 200
		call vardelay           ; perform first 1/5th sec delay
                 
LCDSET          clrf LOOP               ; clr LCD set-up loop
                clrf RSLINE             ; clear RS line for command send
LCDST2          movf LOOP,W             ; get table address
                call TABLCD             ; get set-up instruction
                iorlw 0			; test if zero returned to
		bz   LCDST3		; indicate end of table data
		call LCDOUT             ; not zero. send to lcd
                incf LOOP,F             ; inc loop
                b    LCDST2
LCDST3		movlw 200
		call vardelay           ; yes, perform second 1/5th sec delay
                                        ; to allow final LCD command to occur
                                        ; (it takes longer than the rest)
                 
		return			;end of lcd setup

;**************************************************************
; START OF PROGRAM MAIN CODE
;**************************************************************
Main
		
; The AD9851 initialises in parallel data mode.
; to switch to serial data mode we need to set
; parallel word W0 to xxxxx011 binary. The DDS60
; board has this pattern hard wired on the 8 data
; bits so we simply have to send a clock pulse
; followed by a load pulse to enter serial data mode.

 		bsf porta,3		;set clock bit
		bcf porta,3		;clear clock bit
		bsf porta,2		;set data bit
		bcf porta,2		;clear data bit

; On the DDS60 the nominal 30 MHz crystal is multiplied by 6
; within the AD9851 to give a clock frequency of 180 MHz. The
; deltaphase required for a frequency output of 1 HZ (DELTAHZ) is
; 2^32 / (6 * 30000000) which is 23.860929422222. The program
; allows the value of 30000000 in the above equation to be adjusted
; by the operator to match the actual measured frequency of the
; clock crystal. This will give a slightly different value for
; DELTAHZ and is the software equivalent of a trimmer capacitor
; on the DDS60 board. The adjustment feature is activated by
; keeping the up/down button pressed when the unit is powered up.
; The frequency can be adjusted by means of the 1 Hz and 10 Hz
; buttons in conjunction with the up/down button.
; When the adjusted frequency has been entered, the unit should
; be powered off & on again for the new value to take effect.
;
; Because this software controls the DDS60 in discrete 1 Hz
; steps, we simply need to multiply the DELTAHZ value by the
; required frequency in Hz to obtain the required number to be
; serial loaded into the DDS60.
;
; Note that within the program the DELTAHZ value is scaled by
; a factor of 100 million to give a nominal value of 2386092942,
; this allows the use of fixed point math throughout.
;
; Read the adjusted crystal clock frequency from
; bytes 0, 1, 2 & 3 of EEPROM

		movlw 0			;eeprom address 0
		call geteeprom
		movwf XTAL3
		movlw 1			;eeprom address 1
		call geteeprom
		movwf XTAL2
		movlw 2			;eeprom address 2
		call geteeprom
		movwf XTAL1
		movlw 3			;eeprom address 3
		call geteeprom
		movwf XTAL0	

; Check if up/down button is pressed at switch-on and enter
; adjust mode if so. Adjust mode allows the crystal clock freq
; (default 30 MHz) to be manually adjusted if required.
; This is the software equivalent of adjusting a crystal
; trimming capacitor on the dds unit.
 
		call adjustmode		; adjust nominal clock if
					; reqd by operator		
		call calculatedeltahz	; calulate deltahz value
; Revision 1 of this code set the frequency to 1 MHz
; at switch-on. Revision 2 sets the frequency which
; was set at last switch-off. This is stored in EEprom
; locations 4,5,6 and 7
		call SetInitialFrequency

Mainloop

Check1Hz
		btfsc Portc,0		; 1 Hz button pressed ?
		b Check10Hz		; branch if not
		call beep
		call INCR1Hz  		; increment by 1 Hz
		b MainLoopEnd
Check10Hz
		btfsc Portc,1		; 10 Hz button pressed ?
		b Check100Hz		; branch if not
		call beep
		call INCR10Hz  		; increment by 10 Hz
		b MainLoopEnd
Check100Hz
		btfsc Portc,2		; 100 Hz button pressed ?
		b Check1KHz		; branch if not
		call beep
		call INCR100Hz 		; increment by 100 Hz
		b MainLoopEnd
Check1KHz
		btfsc Portc,3		; 1 KHz button pressed ?
		b Check10KHz		; branch if not
		call beep
		call INCR1KHz 		; increment by 1 KHz
		b MainLoopEnd
Check10KHz
		btfsc Portc,4		; 10 KHz button pressed ?
		b Check100KHz		; branch if not
		call beep
		call INCR10KHz 		; increment by 10 KHz
		b MainLoopEnd
Check100KHz
		btfsc Portc,5		; 100 KHz button pressed ?
		b Check1MHz		; branch if not
		call beep
		call INCR100KHz		; increment by 100 KHz
		b MainLoopEnd
Check1MHz
		btfsc Portc,6		; 1M Hz button pressed ?
		b Check10MHz		; branch if not
		call beep
		call INCR1MHz 		; increment by 1 MHz
		b MainLoopEnd
Check10MHz
		btfsc Portc,7		; 10 MHz button pressed ?
		b MainloopEnd		; branch if not
		call beep
		call INCR10MHz		; increment by 10 MHz

MainLoopEnd
		movlw 200		; delay for debounce
		call vardelay
		goto mainloop           ; Loop checking buttons
                 
;**************************************************************
; OUTPUT CHARACTER IN W TO LCD
;**************************************************************
LCDOUT          movwf STORE             ; temp store data
                movlw 50                ; set minimum time between sending full bytes to
                movwf LOOPA             ; LCD - value of 50 seems OK for this prog with
DELAY           decfsz LOOPA,F          ; XTAL clk of up to 5MHz, possibly 5.5MHz
                goto DELAY              ; keep decrementing LOOPA until zero
                call SENDIT             ; send MSB
                call SENDIT             ; send LSB
                return 
                 
SENDIT          swapf STORE,F           ; swap data nibbles
                movf STORE,W            ; get data byte
                andlw 15                ; get nibble from byte (LSB)
                iorwf RSLINE,W          ; OR the RS bit
                movwf PORTB             ; output the byte
                bsf PORTB,5             ; set E line high
                bcf PORTB,5             ; set E line low
                return 
;**************************************************************
; LCD COMMANDS LINE1, LINE2, CLEAR
;**************************************************************
LCDCLEAR	movlw b'00000001'
		b LCDCMD
LCDLINE1	movlw b'10000000'
		b LCDCMD
LCDLINE2	movlw b'11000000'
LCDCMD		BCF RSLINE,4    ;sets LCD command/line
	        call LCDOUT     ;and outputs cmmand code to LCD
        	BSF RSLINE,4    ;set RS flag
	        movlw 20
		call vardelay
		return
;**************************************************************
; OUTPUT NUMBER OF SPACES SPECIFIED BY W TO LCD 
;**************************************************************                 
LCDspaces
	movwf temp
LCDspaces1
	movlw ' '
	call lcdout
	decf temp,f
	bnz lcdspaces1
	return
; ***************************************************
; PETER HEMSLEY'S 32-BIT MATHS ROUTINES. FOR THIS
; APPLICATION THEY HAVE BEEN UPGRADED TO 64 BITS
; BY TERRY COTTON - ZS1AYJ
; ***************************************************
; SIGNED MULTIPLY REGA * REGB -> REGA
; Return carry set if overflow
; ***************************************************
multiply
	clrf	MTEMP		;Reset sign flag
	call	chksgna		;Make REGA positive
	skpc
	call	chksgnb		;Make REGB positive
	skpnc
	return			;Overflow
	call	movac		;Move REGA to REGC
	call	cleara		;Clear product
	movlw	D'63'		;Loop counter
	movwf	MCOUNT
muloop	call	slac		;Shift left product and multiplicand
	rlf	REGC7,w		;Test MSB of multiplicand
	skpnc			;If multiplicand bit is a 1 then
	call	addba		;add multiplier to product
	skpc			;Check for overflow
	rlf	REGA7,w
	skpnc
	return
	decfsz	MCOUNT,f	;Next
	goto	muloop
	btfsc	MTEMP,0		;Check result sign
	call	negatea		;Negative
	return

; ***************************************************
; SIGNED DIVIDE REGA / REGB -> REGA
; Remainder in REGC
; Return carry set if overflow or division by zero
; ***************************************************
divide	clrf	MTEMP		;Reset sign flag
	movf	REGB0,w		;Trap division by zero
	iorwf	REGB1,w
	iorwf	REGB2,w
	iorwf	REGB3,w
	iorwf	REGB4,w
	iorwf	REGB5,w
	iorwf	REGB6,w
	iorwf	REGB7,w
	sublw	0
	skpc
	call	chksgna		;Make dividend (REGA) positive
	skpc
	call	chksgnb		;Make divisor (REGB) positive
	skpnc
	return			;Overflow
	clrf	REGC0		;Clear remainder
	clrf	REGC1
	clrf	REGC2
	clrf	REGC3
	clrf	REGC4
	clrf	REGC5
	clrf	REGC6
	clrf	REGC7
	movlw	D'64'		;Loop counter
	movwf	MCOUNT
dvloop	call	slac		;Shift dividend (REGA) msb into remainder (REGC)
	movf	REGB7,w		;Test if remainder (REGC) >= divisor (REGB)
	subwf	REGC7,w
	skpz
	goto	dtstgt
	movf	REGB6,w		
	subwf	REGC6,w
	skpz
	goto	dtstgt
	movf	REGB5,w		
	subwf	REGC5,w
	skpz
	goto	dtstgt
	movf	REGB4,w		
	subwf	REGC4,w
	skpz
	goto	dtstgt
	movf	REGB3,w		
	subwf	REGC3,w
	skpz
	goto	dtstgt
	movf	REGB2,w
	subwf	REGC2,w
	skpz
	goto	dtstgt
	movf	REGB1,w
	subwf	REGC1,w
	skpz
	goto	dtstgt
	movf	REGB0,w
	subwf	REGC0,w
dtstgt	skpc			;Carry set if remainder >= divisor
	goto	dremlt
	movf	REGB0,w		;Subtract divisor (REGB) from remainder (REGC)
	subwf	REGC0,f
	movf	REGB1,w
	skpc
	incfsz	REGB1,w
	subwf	REGC1,f
	movf	REGB2,w
	skpc
	incfsz	REGB2,w
	subwf	REGC2,f
	movf	REGB3,w
	skpc
	incfsz	REGB3,w
	subwf	REGC3,f
	movf	REGB4,w
	skpc
	incfsz	REGB4,w
	subwf	REGC4,f
	movf	REGB5,w
	skpc
	incfsz	REGB5,w
	subwf	REGC5,f
	movf	REGB6,w
	skpc
	incfsz	REGB6,w
	subwf	REGC6,f
	movf	REGB7,w
	skpc
	incfsz	REGB7,w
	subwf	REGC7,f
	clrc
	bsf	REGA0,0		;Set quotient bit
dremlt	decfsz	MCOUNT,f	;Next
	goto	dvloop
	btfsc	MTEMP,0		;Check result sign
	call	negatea		;Negative
	return

; ***************************************************
; SIGNED BINARY TO DECIMAL 
; REGA -> DIGITS 1 (MSD) TO 15 (LSD) & DSIGN
; DSIGN = 0 if REGA is positive, FF if negative
; Return carry set if overflow
; Uses FSR register
; ***************************************************
bin2dec
        call    clrdig          ;Clear all digits
        clrf    MTEMP           ;Reset sign flag
	call	chksgna		;Make REGA positive
	skpnc
        goto BLANKIT            ;Overflow
	movlw	D'64'		;Loop counter
	movwf	MCOUNT
b2dloop	rlf	REGA0,f		;Shift msb into carry
	rlf	REGA1,f
	rlf	REGA2,f
	rlf	REGA3,f
	rlf	REGA4,f
	rlf	REGA5,f
	rlf	REGA6,f
	rlf	REGA7,f
	movlw	DIGIT15
	movwf	FSR		;Pointer to digits
	movlw	D'15'		;15 digits to do
	movwf	DCOUNT
adjlp	rlf	INDF,f		;Shift digit and carry 1 bit left
        movlw   -D'10'
	addwf	INDF,w		;Check and adjust for decimal overflow
	skpnc
	movwf	INDF
	decf	FSR,f		;Next digit
	decfsz	DCOUNT,f
	goto	adjlp
	decfsz	MCOUNT,f	;Next bit
	goto	b2dloop
	btfsc	MTEMP,0		;Check sign
	comf	DSIGN,f		;Negative
	clrc
BLANKIT: movlw 48
        iorwf DIGIT1,F
        iorwf DIGIT2,F
        iorwf DIGIT3,F
        iorwf DIGIT4,F
        iorwf DIGIT5,F
        iorwf DIGIT6,F
        iorwf DIGIT7,F
        iorwf DIGIT8,F
        iorwf DIGIT9,F
        iorwf DIGIT10,F
        iorwf DIGIT11,F
        iorwf DIGIT12,F
        iorwf DIGIT13,F
        iorwf DIGIT14,F
        iorwf DIGIT15,F
        movlw 15          ; blank leading zeros
        movwf LOOP
        movlw DIGIT1
        movwf FSR
BLANK:  movf LOOP,W
        xorwf POINT,W
        btfsc STATUS,Z
        return
        movf INDF,W
        andlw 15
        btfss STATUS,Z
        return
        bcf INDF,4
        incf FSR,F
        decfsz LOOP,F
        goto BLANK
        movlw 48
        iorwf DIGIT15,F
        return

; ***************************************************
; Check sign of REGA and convert negative to positive
; Used by multiply, divide, bin2dec
; ***************************************************
chksgna	rlf	REGA7,w
	skpc
	return			;Positive
	call	negatea		;Negative so negate REGA
	return
; ***************************************************
; Negate REGA
; Used by chksgna, multiply, divide, mod, bin2dec, dec2bin
; ***************************************************

negatea	movf	REGA7,w		;Save sign in w
	andlw	0x80

	comf	REGA0,f		;2's complement
	comf	REGA1,f
	comf	REGA2,f
	comf	REGA3,f
	comf	REGA4,f
	comf	REGA5,f
	comf	REGA6,f
	comf	REGA6,f
	incfsz	REGA0,f
	goto	nega1
	incfsz	REGA1,f
	goto	nega1
	incfsz	REGA2,f
	goto	nega1
	incfsz	REGA3,f
	goto	nega1
	incfsz	REGA4,f
	goto	nega1
	incfsz	REGA5,f
	goto	nega1
	incfsz	REGA6,f
	goto	nega1
	incf	REGA7,f
nega1
	incf	MTEMP,f		;flip sign flag
	addwf	REGA5,w		;Return carry set if -2147483648
	return
; ***************************************************
; Set all digits of 15 char buffer to 0h characters
; Used by bin2dec
; ***************************************************
clrdig	clrf	DSIGN
	clrf	DIGIT1
	clrf	DIGIT2
	clrf	DIGIT3
	clrf	DIGIT4
	clrf	DIGIT5
	clrf	DIGIT6
	clrf	DIGIT7
	clrf	DIGIT8
	clrf	DIGIT9
	clrf	DIGIT10
	clrf	DIGIT11
	clrf	DIGIT12
	clrf	DIGIT13
	clrf	DIGIT14
	clrf	DIGIT15
	return

; ***************************************************
; Shift left REGA and REGC
; Used by multiply, divide
; ***************************************************
slac	rlf	REGA0,f
	rlf	REGA1,f
	rlf	REGA2,f
	rlf	REGA3,f
	rlf	REGA4,f
	rlf	REGA5,f
	rlf	REGA6,f
	rlf	REGA7,f
	
	rlf	REGC0,f
	rlf	REGC1,f
	rlf	REGC2,f
	rlf	REGC3,f
	rlf	REGC4,f
	rlf	REGC5,f
	rlf	REGC6,f
	rlf	REGC7,f
	return

; ***************************************************
;Check sign of REGB and negative convert to positive
;Used by multiply, divide, mod
; ***************************************************
chksgnb	rlf	REGB7,w
	skpc
	return			;Positive
	call	negateb		;negative so negate REGB
	return

; ***************************************************
; Negate REGB
; Used by chksgnb, subtract, multiply, divide, mod
; ***************************************************
negateb	movf	REGB5,w		;Save sign in w
	andlw	0x80
	comf	REGB0,f		;2's complement
	comf	REGB1,f
	comf	REGB2,f
	comf	REGB3,f
	comf	REGB4,f
	comf	REGB5,f
	comf	REGB6,f
	comf	REGB7,f
	incfsz	REGB0,f
	goto	negb1
	incfsz	REGB1,f
	goto	negb1
	incfsz	REGB2,f
	goto	negb1
	incfsz	REGB3,f
	goto	negb1
	incfsz	REGB4,f
	goto	negb1
	incfsz	REGB5,f
	goto	negb1
	incfsz	REGB6,f
	goto	negb1
	incf	REGB7,f
negb1
	incf	MTEMP,f		;flip sign flag
	addwf	REGB7,w		;Return carry set if -2147483648
	return
; ***************************************************
; Move REGA to REGC
; used by multiply
; ***************************************************
movac	movf	REGA0,w
	movwf	REGC0
	movf	REGA1,w
	movwf	REGC1
	movf	REGA2,w
	movwf	REGC2
	movf	REGA3,w
	movwf	REGC3
	movf	REGA4,w
	movwf	REGC4
	movf	REGA5,w
	movwf	REGC5
	movf	REGA6,w
	movwf	REGC6
	movf	REGA7,w
	movwf	REGC7
	return
; ***************************************************
; Move REGC to REGA
; used by multiply
; ***************************************************
movca	movf	REGC0,w
	movwf	REGA0
	movf	REGC1,w
	movwf	REGA1
	movf	REGC2,w
	movwf	REGA2
	movf	REGC3,w
	movwf	REGA3
	movf	REGC4,w
	movwf	REGA4
	movf	REGC5,w
	movwf	REGA5
	movf	REGC6,w
	movwf	REGA6
	movf	REGC7,w
	movwf	REGA7
	return

; ***************************************************
; Clear REGA
; Used by multiply, sqrt
; ***************************************************
CLEARA	clrf	REGA0
	clrf	REGA1
	clrf	REGA2
	clrf	REGA3
	clrf	REGA4
	clrf	REGA5
	clrf	REGA6
	clrf	REGA7
	return
; ***************************************************
; Clear REGB
; ***************************************************
CLEARB	clrf	REGB0
	clrf	REGB1
	clrf	REGB2
	clrf	REGB3
	clrf	REGB4
	clrf	REGB5
	clrf	REGB6
	clrf	REGB7
	return

; ***************************************************
; Add REGB to REGA (Unsigned)
; Used by add, multiply,
; ***************************************************
addba	movf	REGB0,w		;Add byte 0
	addwf	REGA0,f
	movf	REGB1,w		;Add byte 1
	skpnc			;No carry_in, so just add
	incfsz	REGB1,w		;Add carry_in to REGB
	addwf	REGA1,f		;Add and propagate carry_out
	movf	REGB2,w		;Add byte 2
	skpnc
	incfsz	REGB2,w
	addwf	REGA2,f
	movf	REGB3,w		;Add byte 3
	skpnc
	incfsz	REGB3,w
	addwf	REGA3,f
	movf	REGB4,w		;Add byte 4
	skpnc
	incfsz	REGB4,w
	addwf	REGA4,f
	movf	REGB5,w		;Add byte 5
	skpnc
	incfsz	REGB5,w
	addwf	REGA5,f
	movf	REGB6,w		;Add byte 6
	skpnc
	incfsz	REGB6,w
	addwf	REGA6,f
	movf	REGB7,w		;Add byte 7
	skpnc
	incfsz	REGB7,w
	addwf	REGA7,f
	return

; ***************************************************
; SIGNED SUBTRACT REGA - REGB -> REGA
; Return carry set if overflow
; ***************************************************
subtract
	call	negateb		;Negate and add
	skpnc
	return			;Overflow

; ***************************************************
; SIGNED ADD REGA + REGB -> REGA
;Return carry set if overflow
; ***************************************************
add	movf	REGA7,w		;Compare signs
	xorwf	REGB7,w
	movwf	MTEMP
	call	addba		;Add REGB to REGA
	clrc			;Check signs
	movf	REGB7,w		;If signs are same
	xorwf	REGA7,w		;so must result sign
	btfss	MTEMP,7		;else overflow
	addlw	0x80
	return

; ***************************************************
; COMPARE 2 NUMBERS IN REGA & REGB
; Return	W=0 and Z flag if equal
; 		W=1 if rega > reg b
;		W=2 if regb > rega
; ***************************************************
compare
	call subtract
	movf rega0,w	;check if rega = 0
	iorwf rega1
	iorwf rega2
	iorwf rega3
	iorwf rega4
	iorwf rega5
	iorwf rega5
	iorwf rega7
	bnz compare1
	return		;operands are equal. W=0 and Z flag
compare1
	btfss rega7,7
	b compare2
	movlw 2
	return
compare2
	movlw 1
	return

; ***************************************************
; DIVIDE REGA BY 100 MILLION
; ***************************************************
divide100million
	clrf regb7
	clrf regb6
	clrf regb5
	clrf regb4
	movlw 5h
	movwf regb3
	movlw 0f5h
	movwf regb2
	movlw 0e1h
	movwf regb1
	movlw 0h
	movwf regb0
	call divide
	return
; ***************************************************
; MOVE FREQUENCY TO REGA
; ***************************************************
freq2rega
        movf freq0,W    
        movwf REGA0
        movf freq1,W    
        movwf REGA1
        movf freq2,W    
        movwf REGA2
        movf freq3,W    
        movwf REGA3
        clrf REGA4
	clrf REGA5
	clrf REGA6
	clrf REGA7
	return
; ***************************************************
; MOVE REGA TO FREQUENCY
; ***************************************************
rega2freq
        movf rega0,W    
        movwf freq0
        movf rega1,W    
        movwf freq1
        movf rega2,W    
        movwf freq2
        movf rega3,W    
        movwf freq3
	return
; ***************************************************
; WORK REG TO REGB
; ***************************************************
Work2Regb
	call CLEARB
	movwf REGB0
	return

; ***************************************************
; MOVE DELTAHZ VALUE TO REGB
; ***************************************************
DELTAHZ2RegB
	clrf regb7
	clrf regb6
	clrf regb5
	clrf regb4
	movf DELTAHZ3,w
	movwf regb3	
	movf  DELTAHZ2,w
	movwf regb2	
	movf  DELTAHZ1,w
	movwf regb1	
	movf DELTAHZ0,w
	movwf regb0	
	return
; ***************************************************
; CALCULATE THE PHASE REGISTER VALUE WHICH CORRESPONDS
; TO THE FREQUENCY IN THE FREQREG. WHEN CALCULATED THIS
; WILL BE SENT TO THE DDS60 TO CHANGE FREQUENCY
; ***************************************************
SetFreq
	call freq2rega		; Freq to RegA
	call DELTAHZ2regb	; DELTAHZ to regb
	call multiply		; multiply to get phase acc value
	call divide100million	; the DELTAHZ value was scaled
				; by a factor of 100 million so
				; divide by same factor
	call Send2DDS60		; send the new deltaphase value
				; to the DDS60
	call displayfreq	; display the new frequency on lcd
; REVISION 2 - write the frequency to bytes 4-7 of eeprom for the
; switch-on initialisation routine
	movf freq3,w
	movwf eepromdata
	movlw 4
	call seteeprom
	movf freq2,w
	movwf eepromdata
	movlw 5
	call seteeprom
	movf freq1,w
	movwf eepromdata
	movlw 6
	call seteeprom
	movf freq0,w
	movwf eepromdata
	movlw 7
	call seteeprom
	return

; ***************************************************
; SEND THE DELTAPHASE VALUE IN REGA TO THE DDS60
; ***************************************************
Send2DDS60
	movf rega0,w	; send 4 deltaphase bytes in least
	call sendbyte	; significant bit to most significant
	movf rega1,w	; bit order
	call sendbyte
	movf rega2,w
	call sendbyte
	movf rega3,w
	call sendbyte
	movlw 09h	; the last 8 bits of the 40 bit serial
	call sendbyte	; data stream are preset to 09h which
			; indicates phase=0, power-up and
			; 6 x Clock enabled
	bsf porta,2	; 40 bits have been sent. Now send
	bcf porta,2	; a load pulse to activate new frequency
	return
; ***************************************************
; SEND THE BYTE IN W TO THE DDS60 LSB TO MSB
; ***************************************************
SendByte
	movwf temp	;save the byte to be sent
	movlw 8
	movwf loop	;send 8 bits
Sendbyte1
	movf temp,w
	andlw 1		;extract ls bit
	bz SendByte2
	bsf porta,4	;set data bit
	b SendByte3
SendByte2
	bcf porta,4	;reset data bit
SendByte3
	bsf porta,3	;set clock bit
	bcf porta,3	;clear clock bit
	rrf temp,f	;shift 1 bit right
	decf loop,f
	bnz SendByte1
	return
; ***************************************************
; DISPLAY THE FREQUENCY HELD IN THE FREQREG
; ON TOP LINE OF LCD
; ***************************************************
DisplayFreq
	call freq2rega		; convert freq to decimal
	call bin2dec
	call lcdline1
	movlw 2 		; 2 leading spaces
	call lcdspaces
	movf digit8,w
	call lcdout
	movf digit9,w		; output digits to lcd
	call lcdout
	movlw 20h		; if digit 9 isn't a space
	subwf digit9,w		; then output a comma as
	bz displayfreq1		; a million separator
	movlw ','
	call lcdout
displayfreq1
	movf digit10,w
	call lcdout
	movf digit11,w
	call lcdout
	movf digit12,w
	call lcdout
	movlw ','		; thousand separator
	call lcdout
	movf digit13,w
	call lcdout
	movf digit14,w
	call lcdout
	movf digit15,w
	call lcdout
	movlw ' '
	call lcdout
	movlw 'H'
	call lcdout
	movlw 'z'
	call lcdout
	movlw 3 		; 3 trailing spaces
	call lcdspaces
	return
; ***************************************************
; DELAY APPROXIMATELY 1 mS WITH 4 MHZ CLOCK
; ***************************************************
msdelay
	movlw 248
	movwf msloop
msdelay1
	nop
	decfsz msloop
	b msdelay1
	nop
	nop
	nop
	return
; ***************************************************
; DELAY APPROXIMATELY 1 mS x THE VALUE IN W
; ***************************************************
VarDelay
	movwf varloop
Vardelay1
	call msdelay
	decfsz varloop
	b vardelay1
	return
	
; ***************************************************
; INCREASE OR DECREASE FREQ BY 1 HZ
; ***************************************************
INCR1Hz
	call CLEARB
	movlw 1			
	movwf regb0		
	call changefreq		
	return
; ***************************************************
; INCREASE OR DECREASE FREQ BY 10 HZ
; ***************************************************
INCR10Hz
	call CLEARB
	movlw 0Ah		
	movwf regb0		
	call changefreq		
	return
; ***************************************************
; INCREASE OR DECREASE FREQ BY 100 HZ
; ***************************************************
INCR100Hz
	call CLEARB
	movlw 064h
	movwf regb0
	call changefreq		 
	return

; ***************************************************
; INCREASE OR DECREASE FREQ BY 1 KHZ
; ***************************************************
INCR1KHz
	call CLEARB
	movlw 03h		
	movwf regb1		
	movlw 0E8h
	movwf regb0
	call changefreq		
	return
; ***************************************************
; INCREASE OR DECREASE FREQ BY 10 KHZ
; ***************************************************
INCR10KHz
	call CLEARB
	movlw 027h		
	movwf regb1		
	movlw 010h
	movwf regb0
	call changefreq		
	return
; ***************************************************
; INCREASE OR DECREASE FREQ BY 100 KHZ
; ***************************************************
INCR100KHz
	call CLEARB
	movlw 01h		
	movwf regb2		
	movlw 086h		
	movwf regb1		
	movlw 0A0h
	movwf regb0
	call changefreq		
	return
; ***************************************************
; INCREASE OR DECREASE FREQ BY 1 MHZ
; ***************************************************
INCR1MHz
	call CLEARB
	movlw 0Fh		
	movwf regb2		
	movlw 042h		
	movwf regb1		
	movlw 040h
	movwf regb0
	call changefreq		
	return
; ***************************************************
; INCREASE OR DECREASE FREQ BY 10 MHZ
; ***************************************************
INCR10MHz
	call CLEARB
	movlw 098h		
	movwf regb2		
	movlw 096h		
	movwf regb1		
	movlw 080h
	movwf regb0
	call changefreq		
	return
; ***************************************************
; INCREASE OR DECREASE FREQ BY VALUE IN REG B
; ***************************************************
ChangeFreq
	call freq2rega		; move freq reg to rega
	btfss porta,0		; check if up/down button pressed
	b ChangeFreq1		; branch if pressed
	call addba		; up/down button not pressed
	b changefreq2		; so increment frequency
ChangeFreq1
	call subtract		; up/down button pressed
	b changefreq2		; so decrement frequency
Changefreq2
	Call AdjustLimits       ; Check if freq outside limits
	call rega2freq		; and adjust if neccessary
	call setfreq		; send frequency change to dds
	return
; ***************************************************
; CHECK IF REGA IS WITHIN FREQUENCY LIMITS
; HIGHLIMIT to LOWLIMIT AND ADJUST IF NECCESSARY
; ***************************************************
AdjustLimits
	call movac		; preserve rega
	call HIGHLIMIT2B	; 60 Million (60 MHz) to regb
	incf regb0,f		; bump by 1 for upper limit comparison
	call subtract
	btfss rega5,7
	b FreqTooHigh		; branch if above upper limit
	call movca		; frequency is not above the upper limit
	call LOWLIMIT2B		; move low freq limit to regb
	call subtract		; subtract from freq & check if negative
	btfsc rega5,7
	b FreqTooLow		; branch if negative (too low)
	call movca
	return
FreqTooHigh
	call movca		; restore freq to rega
	call HIGHLIMIT2B
	call subtract
	return
FreqTooLow
	call movca		; restore freq to rega
	call HIGHLIMIT2B	; 60 million (60 Mhz) to regb
	call addba
	return
; ***************************************************
; MOVE HIGH LIMIT FOR FREQUENCY INTO REGB THIS IS SET
; TO 60 MHz (03938700h) BUT CAN BE CHANGED TO ANY OTHER VALUE
; ***************************************************
HIGHLIMIT2B	
	call clearb		 
	movlw 03h		
	movwf regb3
	movlw 093h		
	movwf regb2
	movlw 087h
	movwf regb1
	movlw 00h
	movwf regb0
	return
; ***************************************************
; MOVE LOW LIMIT FOR FREQUENCY INTO REGB THIS IS SET
; TO 10 KHz (2710h) BUT CAN BE CHANGED TO ANY OTHER VALUE
; THE AD9851 HAS NO LOW LIMIT BUT THE COUPLING CAPACITORS
; TO THE WIDEBAND AMPLIFIER CAUSE A RAPID DROP OFF
; IN LF RESPONSE BELOW 10 KHz 
; ***************************************************
LOWLIMIT2B	
	call clearb		 
	movlw 027h
	movwf regb1
	movlw 010h
	movwf regb0
	return
; ***************************************************
; BEEP ON EACH BUTTON PUSH. PULSES A SMALL PIEZO
; SOUNDER CONNECTED TO RA1
; ***************************************************

BEEP		movlw 255
		movwf beepouter
beep1		bsf porta,1
		call beepinnerdelay
		bcf porta,1
		call beepinnerdelay
		decf beepouter,f
		bnz beep1
		return
beepinnerdelay
		movlw 	45
		movwf	beepinner
beepinnerdelay1
		decf beepinner,f
		bnz beepinnerdelay1
		return

; ***************************************************
; READ DATA FROM EEPROM ROUTINE FOR PIC16F87x DEVICES
; ENTER WITH EEPROM ADDRESS IN W
; RETURNS WITH DATA IN W
; ***************************************************
GETEEPROM
	BANK2
        movwf EEADR     	;copy W into EEADR to set eeprom address
	BANK3
        bcf EECON1,EEPGD 	;point to data memory
        bsf EECON1,RD   	;enable read flag
	BANK2
        movf EEDATA,W   	;read eeprom data now in EEDATA into W
	BANK0
        return

; ***************************************************
; WRITE DATA TO EEPROM ROUTINE FOR PIC16F87x DEVICES
; ENTER WITH EEPROM ADDRESS IN W AND DATA TO BE
; WRITTEN IN VARIABLE EEPROMDATA
; ***************************************************
SETEEPROM
	BANK2
        movwf EEADR     	;copy W into EEADR to set eeprom address
	BANK0
        MOVF EEPROMDATA,W  	;get data value from STORE1 and hold in W
	BANK2
        movwf EEDATA    	;copy W into eeprom data byte register
	BANK3
        bcf EECON1,EEPGD 	;point to Data memory
        bsf EECON1,WREN 	;enable write flag
MANUAL  movlw 55h       	;these lines cause the action required by
        movwf EECON2    	;by the eeprom to store the data in EEDATA
        movlw 0AAh       	;at the address held by EEADR.
        movwf EECON2
        bsf EECON1,WR   	;set the ``perform write'' flag
	BANK0
CHKWRT  btfss PIR2,EEIF 	;wait until bit 4 of PIR2 is set
        goto CHKWRT
        bcf PIR2,EEIF   	;clear bit 4 of PIR2
        return
;**************************************************************
; Show nominal DDS60 crystal clock frequency on LCD line 1
;**************************************************************                 
displaynominal
	call lcdline1
	movlw 'N'
	call lcdout        
	movlw 'o'
	call lcdout        
	movlw 'm'
	call lcdout        
	movlw 'i'
	call lcdout        
	movlw 'n'
	call lcdout        
	movlw 'a'
	call lcdout        
	movlw 'l'
	call lcdout        
	movlw 1
	call lcdspaces
	movlw '3'
	call lcdout        
	movlw '0'
	call lcdout        
	movlw '0'
	call lcdout        
	movlw '0'
	call lcdout        
	movlw '0'
	call lcdout        
	movlw '0'
	call lcdout        
	movlw '0'
	call lcdout        
	movlw '0'
	call lcdout        
	return
;**************************************************************
; Show adjusted DDS60 crystal clock frequency on LCD line 2
;**************************************************************                 
displayadjusted
	call lcdline2
	movlw 'A'
	call lcdout        
	movlw 'd'
	call lcdout        
	movlw 'j'
	call lcdout        
	movlw 'u'
	call lcdout        
	movlw 's'
	call lcdout        
	movlw 't'
	call lcdout        
	movlw 2
	call lcdspaces
	call cleara		; move xtal value to rega
	movf XTAL3,w
	movwf rega3
	movf XTAL2,w
	movwf rega2
	movf XTAL1,w
	movwf rega1
	movf XTAL0,w
	movwf rega0
	call bin2dec		; convert to decimal
	movf digit8,w
	call lcdout
	movf digit9,w
	call lcdout
	movf digit10,w
	call lcdout
	movf digit11,w
	call lcdout
	movf digit12,w
	call lcdout
	movf digit13,w
	call lcdout
	movf digit14,w
	call lcdout
	movf digit15,w
	call lcdout
	return
; ***************************************************
; CHECK IF UP/DOWN BUTTON IS PRESSED AT SWITCH-ON
; AND ENTER ADJUST MODE IF SO
; ***************************************************
AdjustMode
		btfsc Porta,0	; Up/down button pressed ?
		return		; exit if not
		call beep
		call displaynominal ; display nominal clock crystal freq
		call displayadjusted ; display adjusted frequency

AdjustCheck1Hz
		movlw 1
		movwf temp
		btfsc Portc,0		; 1 Hz button pressed ?
		b AdjustCheck10Hz	; branch if not
		call AdjustFreq		; bump + - 1 Hz
		b AdjustLoopEnd
AdjustCheck10Hz
		movlw 10
		movwf temp
		btfsc Portc,1		; 10 Hz button pressed ?
		b AdjustCheck1Hz	; branch if not
		call AdjustFreq		; bump + - 10 Hz
AdjustLoopEnd
		movlw 200		; debounce delay
		call vardelay
		b AdjustCheck1Hz	; loop until switch-off
AdjustFreq		
		call beep
		call cleara	; clear rega
		call clearb	; set regb = 1 or 10
		movf temp,w
		movwf regb0
		btfss porta,0	; check if up/down button pressed
		b AdjustFreq2	; branch if pressed
		call addba	; up/down button not pressed so add
				; regb (1 or 10) to rega (0) to
				; give rega = 1 or 10
		b AdjustFreq3	; so increment frequency
AdjustFreq2
		call subtract	; up/down button pressed so subtract
				; regb (1 or 10) from rega (0) to
				; give rega = -1 or -10
AdjustFreq3	call clearb	; xtal frequency to regb
		movf xtal3,w
		movwf regb3
		movf xtal2,w
		movwf regb2
		movf xtal1,w
		movwf regb1
		movf xtal0,w
		movwf regb0
		call addba	; add the + - 1 or 10 to phase increment
		movf rega3,w	; move rega back as the new
		movwf xtal3	; adjusted crystal frequency
		movf rega2,w	
		movwf xtal2 
		movf rega1,w	
		movwf xtal1 
		movf rega0,w	
		movwf xtal0 
		call displayadjusted ; display adjusted freq
		movf xtal3,w ; write new adjusted freq to eeprom
		movwf eepromdata
		movlw 0
		call seteeprom
		movf xtal2,w
		movwf eepromdata
		movlw 1
		call seteeprom
		movf xtal1,w
		movwf eepromdata
		movlw 2
		call seteeprom
		movf xtal0,w
		movwf eepromdata
		movlw 3
		call seteeprom
		return
; ***************************************************
; CALCULATE THE DELTAHZ VALUE (2^32 * 10^8) / (6 * CLKadj)
; ***************************************************
CalculateDeltaHz
	call cleara	; move adjusted crystal clock value
	movf xtal0,w	; to rega (30000000 nominal)
	movwf rega0
	movf xtal1,w
	movwf rega1
	movf xtal2,w
	movwf rega2
	movf xtal3,w
	movwf rega3
	call clearb	; 6 to regb
	movlw 6
	movwf regb0
	call multiply	; rega contains 6 * CLKadj 180 MHz nominal
	call clearb	; rega to regb
	movf rega0,w
	movwf regb0
	movf rega1,w
	movwf regb1
	movf rega2,w
	movwf regb2
	movf rega3,w
	movwf regb3
	call cleara	; move the value 2^32 scaled by 100 million
	movlw 5		; to rega. This value is
	movwf rega7	; 05 F5 E1 00 00 00 00 00 HEX
	movlw 0F5h
	movwf rega6
	movlw 0E1h
	movwf rega5
	call divide	; (2^32 * 10^8)/ (6 * CLKadj)
	movf rega3,w	; save calculated value of deltahz
	movwf deltahz3
	movf rega2,w	
	movwf deltahz2
	movf rega1,w	
	movwf deltahz1
	movf rega0,w	
	movwf deltahz0
	return
	
; ***************************************************
; REVISION 2
; If the contents of EEprom locations 4,5,6 and 7 are
; all zero then this is the first switch-on since the
; device was programmed so set the initial frequency
; to 1 MHz
; Otherwise set the frequency to that in EEprom bytes
; 4-7 which is the frequency at last switch-off 
; ***************************************************
SetInitialFrequency
		clrf temp
		movlw 4
		call GetEEprom 
		movwf freq3
		iorwf temp,temp
		movlw 5
		call GetEEprom 
		movwf freq2
		iorwf temp,temp
		movlw 6
		call GetEEprom 
		movwf freq1
		iorwf temp,temp
		movlw 7
		call GetEEprom 
		movwf freq0
		iorwf temp,temp		; temp will be zero if first switch-on
		bz Set1MHz			; so go to set freq to 1 MHx



		Call Setfreq	; set frequency to that of last switch-off
		return
Set1MHz
		clrf freq3
		movlw 0fh		; set freq register to 1 MHz
		movwf freq2		; i.e 1,000,000 Hz which is
		movlw 042h		; hex F4240
		movwf freq1		
		movlw 040h
		movwf freq0
		call SetFreq	; set frequency to 1 MHz
		return

crash	b crash

		org 2100h
NominalXtal
; Nominal frequency of the 30 MHz Xtal oscillator module.
; This can be modified for fine tuning by holding the up/down button
; pressed at switch-on. The value is modified with the 1 Hz button
; in conjunction with up/down. The new value is stored in EEPROM
; Switch off & on again for normal operation using the new value
		DE 1,0C9h,0C3h,80h	

; REVISION 2
; Set bytes 4-7 of EEprom to zero. At first switch-on this
; will indicate that the frequency should be initialised to 1 MHz.
; On subsequent switch-ons these 4 bytes will contain the
; frequency at the last switch-off and thae frequency will
; be initialised to this value.
		De 0,0,0,0
	end 
